-- * All helper identifiers (either commands and variable bindings) start with
--   ġė_ prefix to distinguish ghciExt related stuff from user identifiers.
-- * Final commands do not have the prefix and are registered in :defs command.
-- * Commands which store information into the top level bindings or into :def
--   command cannot be called recursively (the scopes are global).

-- just my defaults (ghciExt does not depend on these)
:set -fbreak-on-error
:set -XLambdaCase
-- :set -fbreak-on-exception
:set prompt "\ESC[1;32m\STX%s>\ESC[0m\STX "

-- these are important for ghciExt to work correctly
:set stop
:set -fno-print-bind-result

-- :init command defines readsBase :: Num a => a -> String -> [(a, String)]
-- Redefine :init to reintroduce any other top level symbols which
-- GHCi clears when your program is running.
:{
:def init const $ return $ concat
     ["let{"
     ,"readsAtBase b s ="
     ," let c2i c ="
     ,"  let i = fromEnum c in"
     ,"  if i>=97 then i-87 else if i>=65 then i-55 else i-48 in"
     ," let isValid c = Data.Char.isAlphaNum c && c2i c < b in"
     ," Numeric.readInt b isValid c2i s ;"
     ,"showsAtBase b i ="
     ," let i2c x = if x<10 then toEnum(48+x) else toEnum(55+x) in"
     ," Numeric.showIntAtBase b i2c i }" ]
:}
:init

------------------------------------------------------------------------------
-- Simple redefines of the builtin commands:

-- :back, :forward, :step, :steplocal, :stepmodule should :list too
:def back const $ return ":redir ġė_Tmp ::back\nputStrLn (head $ lines ġė_Tmp)\nlet ġė_Tmp=()\n:list"
:def forward const $ return ":redir ġė_Tmp ::forward\nputStrLn (head $ lines ġė_Tmp)\nlet ġė_Tmp=()\n:list"
:def step const $ return "::step\n:list\n:init"
:def steplocal const $ return "::steplocal\n:list\n:init"
:def stepmodule const $ return "::stepmodule\n:list\n:init"

-- :break should support breakpoint hit count
:def break \args -> return $":ġė_Inject Breaking: True \"\" True "++args

-- do not show internal variables (ġė_*) in bindings by default
:{
:def show \keyword ->
     if keyword /= "bindings" then return $ "::show "++keyword else
     return $ unlines
       [":redir ġė_Text ::show bindings"
       ,"putStr $ unlines $ filter (not . Data.List.isPrefixOf \"ġė_\") $ lines ġė_Text" ]
:}

-- :force should work with an expression by default
:{
:def force \expr -> return $ unlines
     ["let ġė_Tmp = "++expr
     ,":redir ġė_Text ::force ġė_Tmp"
     ,"let ġė_Tmp = ()" {- keep show bindings terse -}
     ,"let{ dropVarName l = let p = \"ġė_Tmp = \" in " ++
       "if Data.List.isPrefixOf p l then drop (length p) l else l }in " ++
       "mapM_ (putStrLn . dropVarName) $lines ġė_Text"
     ,"let ġė_Text = ()" ] {- keep show bindings terse -}
:}

------------------------------------------------------------------------------

ġė_NullFileHnd <- System.IO.openFile "/dev/null" System.IO.WriteMode
ġė_OrgGhcStdOut <- GHC.IO.Handle.hDuplicate System.IO.stdout
ġė_OrgGhcStdErr <- GHC.IO.Handle.hDuplicate System.IO.stderr

:{
:def ġė_OutOn
     ( const ( GHC.IO.Handle.hDuplicateTo ġė_OrgGhcStdOut System.IO.stdout >> return "" ) )
:}

:{
:def ġė_OutOff
     ( const ( GHC.IO.Handle.hDuplicateTo ġė_NullFileHnd System.IO.stdout >> return "" ) )
:}

:{
:def ġė_ErrOn
     ( const ( GHC.IO.Handle.hDuplicateTo ġė_OrgGhcStdErr System.IO.stderr >> return "" ) )
:}

:{
:def ġė_ErrOff
     ( const ( GHC.IO.Handle.hDuplicateTo ġė_NullFileHnd System.IO.stderr >> return "" ) )
:}

:{
:def output GhciExt.cmdHlp
     ( const ( GHC.IO.Handle.hDuplicateTo ġė_OrgGhcStdOut System.IO.stdout >>
               GHC.IO.Handle.hDuplicateTo ġė_OrgGhcStdErr System.IO.stderr >>
               return "" ) )
     ":output                   -- restore ghci output (stdout and stderr) back"
     $unlines $map (\s->' ':s)
       ["Restores stdout and stderr of ghci as it was during startup. You should not"
       ,"ever need this command. It is here only if something would go wrong in ghciExt"
       ,"and you need to recover (console) manually." ]
:}

:{
:def . GhciExt.cmdHlp
     (return . (":script "++))
     ":. <file>                 -- source commands from <file>" ""
:}

:{
:def * GhciExt.cmdHlp
     ( \cntCmd ->
       case break Data.Char.isSpace cntCmd of {
         ( cnt, _:cmd ) -> return $unlines $replicate (read cnt) cmd ;
         _ -> return "putStrLn \"usage: :* <count> <cmd>...\"" } )
     ":* <count> <cmd>...       -- run <cmd> <count> times" ""
:}

:{
:def x GhciExt.cmdHlp
     ( \cmdAndItsArgs -> return $ unlines
        [":ġė_OutOff"
        ,cmdAndItsArgs
        ,":ġė_OutOn"] )
     ":x <cmd>...               -- run <cmd> with stdout suppressed" ""
:}

:{
:def X GhciExt.cmdHlp
     ( \cmdAndItsArgs -> return $ unlines
        [":ġė_OutOff"
        ,":ġė_ErrOff"
        ,cmdAndItsArgs
        ,":ġė_ErrOn"
        ,":ġė_OutOn"] )
     ":X <cmd>...               -- run <cmd> with stdout and stderr suppressed" ""
:}

:{
:def redir GhciExt.cmdHlp
     ( \varCmd ->
       case break Data.Char.isSpace varCmd of {
         (var,_:cmd) -> return $ unlines
           ["ġė_RedirInfo <- GhciExt.setOutToTmpFile"
           ,":def ġė_RedirRestore const ( GhciExt.restoreOutGetTmpFileContent ġė_RedirInfo"++
             " >>= \\t -> return $ \""++var++" <- return \"++show t )"
           ,"let ġė_RedirInfo = ()" {- keep show bindings terse -}
           , cmd
           ,":ġė_RedirRestore"
           ,":undef ġė_RedirRestore" ] ;
         _ -> return "putStrLn \"usage: :redir <var> <cmd>...\"" } )
     ":redir <var> <cmd>...     -- execute <cmd> redirecting stdout to <var>" ""
:}

:{
:def grep GhciExt.cmdHlp
     ( \patCmd ->
       case GhciExt.extractArgUpToDelimiter '<' patCmd of {
         [grepArgs,cmd] -> return $ unlines
           ["ġė_RedirInfo <- GhciExt.setOutToTmpFile"
           ,":def ġė_GrepRestore const ( GhciExt.restoreOutGetTmpFileName ġė_RedirInfo"++
             " >>= GhciExt.executeGrep "++show grepArgs++show cmd++" >> return \"\" )"
           ,"let ġė_RedirInfo = ()" {- keep show bindings terse -}
           , cmd
           ,":ġė_GrepRestore"
           ,":undef ġė_GrepRestore" ] ;
         _ -> return "putStrLn \"usage: :grep <args> < <cmd>...\"" } )
     ":grep <args> < <cmd>...   -- run grep <args> on <cmd> output"
     $unlines $map (\s->' ':s)
       ["Runs grep from your OS on the output of <cmd>. This means all the options"
       ,"of your OS grep are available. '<' separates grep options from the command." ]
:}

:{
:def ġė_Inject
     ( \args ->
       let [msg,codeCond,code,stopCond,bpArgs] = GhciExt.parseStrArgs 4 args in
         return $ unlines
           [":redir ġė_Text ::break "++bpArgs
           ,"putStr $ "++show msg++
             "++' ':GhciExt.highlightIf False \"^Breakpoint [0-9]+ activated at \" ġė_Text"
           ,":cmd return$GhciExt.getSetStopCode ġė_Text "++
             show codeCond++" "++ show code++" "++show stopCond
           ,"let ġė_Text = ()" {- keep show bindings terse -} ] )
:}

:{
:def inject GhciExt.cmdHlp
     ( \args ->
       case GhciExt.parseStrArgs 3 args of {
         [codeCond,code,stopCond,bpArgs] ->
           return $":ġė_Inject Injecting: "++show codeCond++show code++show stopCond++bpArgs ;
         _ -> return "putStrLn \"usage: :inject <codeCond> <code> <stopCond> <bpArgs>\"" } )
     ":inject <cc> <c> <sc> <b> -- at location <b> run <c> if <cc> and stop if <sc>"
     $unlines $map (\s->' ':s)
       ["There are two special identifiers which can be used in the breakpoint"
       ,"code (<c>) and the breakpoint stop condition (<sc>); but which are not"
       ,"usable in the breakpoint code condition (<cc>). These are:"
       ," - ġė_BpId - macro name; translates to breakpoint id"
       ," - ġė_HitCnt - macro name; translates to breakpoint hit count" ]
:}

:{
:def strobe GhciExt.cmdHlp
     ( \rawArgs ->
       let args = case rawArgs of{ '&':rawArgsTail -> rawArgsTail; _ -> "\"\" "++rawArgs } in
       case GhciExt.parseStrArgs 1 args of {
         [cond,bpArgs] ->
           let code = "putStrLn $\"ġė_BpId: HIT \"++show ġė_HitCnt" in
           return $":ġė_Inject Strobing: "++show cond++show code++" False "++bpArgs ;
         _ -> return "putStrLn \"usage: :strobe [&<cond>] <bpArgs>\"" } )
     ":strobe [&<c>] <b>        -- at location <b> show hit count if <c>" ""
:}

:{
:def monitor GhciExt.cmdHlp
     ( \rawArgs ->
       let args = case rawArgs of{ '&':rawArgsTail -> rawArgsTail; _ -> "\"\" "++rawArgs } in
       case GhciExt.parseStrArgs 2 args of {
         [cond,expr,bpArgs] ->
           let code = "putStr \"ġė_BpId: \"\n"++expr in
           return $":ġė_Inject Monitoring: "++show cond++show code++"False "++bpArgs ;
         _ -> return "putStrLn \"usage: :monitor [&<cond>] <expr> <bpArgs>\"" } )
     ":monitor [&<c>] <e> <b>   -- print expression <e> at location <b> if <c>" ""
:}

:{
:def watch GhciExt.cmdHlp
     ( \rawArgs ->
       let args = case rawArgs of{ '&':rawArgsTail -> rawArgsTail; _ -> "\"\" "++rawArgs } in
       case GhciExt.parseStrArgs 2 args of {
         [cond,stopCond,bpArgs] ->
           return $":ġė_Inject Watching: "++show cond++" \"\" "++show stopCond++bpArgs ;
         _ -> return "putStrLn \"usage: :watch [&<cond>] <stopCond> <bpArgs>\"" } )
     ":watch [&<c>] <sc> <b>    -- if <c> then break at location <b> when <sc> is True" ""
:}

:{
:def count GhciExt.cmdHlp
     ( \rawArgs ->
       let argParseRv2Touple rv = case rv of {[a]->(a,""); [a,b]->(a,b); _->("","")} in
       let{ (cond,relOpCntBpArgs) = case rawArgs of{
         '&':rawArgsTail -> argParseRv2Touple $ GhciExt.parseStrArgs 1 rawArgsTail;
         _ -> ("", rawArgs) } }in
       let (relOpCnt,bpArgs) = argParseRv2Touple $ GhciExt.parseStrArgs 1 relOpCntBpArgs in
       let{ (relOp,cnt) = case relOpCnt of { '/':'=':r->("/=",r); '=':'=':r->("==",r);
         '<':'=':r->("<=",r); '>':'=':r->(">=",r); '<':r->("<",r); '>':r->(">",r); r->("==",r) } }in
       if not (null cnt) && all Data.Char.isDigit cnt
         then {- command arguments seem to be correct -}
           if null cond && null bpArgs then {- query hit count of a breakpoint with given id -}
             return $ unlines
               [":ġė_HitCnt"++cnt
               ,"putStr $ \"Hit count for breakpoint "++cnt++" is \""
               ,"ġė_HitCnt"++cnt ]
           else if relOp=="==" && cnt=="0" then {- setup a breakpoint just to count hits -}
             return $ ":ġė_Inject Counting: "++show cond++" \"\" False "++bpArgs
           else {- setup counting breakpoint to stop at specified hit -}
             return $ ":ġė_Inject Counting: "++show cond++" \"\" ġė_HitCnt"++relOp++cnt++" "++bpArgs
         else return "putStrLn \"usage: :count [&<cond>] [<relOp>]<Num> [<bpArgs>]\"" )
     ":count [&<c>] [<R>]<N> [<bp>] -- count/stop execution at <bp> or query bp hits"
     $unlines $map (\s->' ':s)
       ["There are four ways how to use this command:"
       ," - :count 0 <bp>  - never stops; counts hits and extends trace history"
       ," - :count N <bp>  - stops when location <bp> is hit N-th time"
       ," - :count RN <bp> - stops when hitCount@bp is in Ord relation R with N"
       ," - :count N  - shows hit count of a ghciExt breakpoint with number N."
       ,"If &<c> is there then <c> limits whether the above runs at all." ]
:}

:{
:def find GhciExt.cmdHlp
     ( \cmdVar ->
       case GhciExt.parseStrArgs 1 cmdVar of {
         [cmd, var'] | (not $null cmd) && (not $null var') ->
           let (patBody,var) = if '-' == head var' then (tail var', "") else (var', var') in
           return $ ":ġė_Search \"\\n"++patBody++" :: \""++show var++cmd ;
         _ -> return "putStrLn \"usage: :find <cmd> <var>\"" } )
     ":find <cmd> [-]<var>      -- step with <cmd> until <var> is found"
     " Prepend variable name with '-' to disable printing of its value.\n"
:}

:{
:def locate GhciExt.cmdHlp
     ( \cmdBpArgs ->
       case GhciExt.parseStrArgs 1 cmdBpArgs of {
         [cmd, bpArgs] | (not $null cmd) -> return $unlines
           [":redir ġė_Text ::break "++bpArgs
           ,":cmd return$GhciExt.getLocateCode ġė_Text "++show cmd
           ,"let ġė_Text = ()" {- keep show bindings terse -} ] ;
         _ -> return "putStrLn \"usage: :locate <cmd> <bpArgs>\"" } )
     ":locate <cmd> <bpArgs>    -- step with <cmd> until location <bpArgs> is hit" ""
:}

:{
:def ġė_Search
     ( \args ->
       let [pat,var,cmd] = GhciExt.parseStrArgs 2 args in return $ unlines
         [":redir ġė_Text "++cmd
         ,":cmd return$GhciExt.getSearchCode ġė_Text "++show pat++show var++show cmd
         ,"let ġė_Text = ()" {- keep show bindings terse -} ] )
:}

:{
:def repeat GhciExt.cmdHlp
     ( \cmdBpArgs ->
       case GhciExt.parseStrArgs 1 cmdBpArgs of {
         [cmd, expr] | (not $null cmd) -> return $":ġė_Repeat "++show expr++cmd;
         _ -> return "putStrLn \"usage: :repeat <cmd> <cond>\"" } )
     ":repeat <cmd> <cond>      -- repeat <cmd> until <cond> is true" ""
:}

:{
:def ġė_Repeat
     ( \args ->
       let [expr,cmd] = GhciExt.parseStrArgs 1 args in return $ unlines
         [":redir ġė_Text "++cmd
         ,":ġė_ErrOff"
         ,":redir ġė_Tmp "++expr
         ,":ġė_ErrOn"
         ,":cmd return$GhciExt.getRepeatCode ġė_Text ġė_Tmp "++show expr++show cmd
         ,"let ġė_Tmp = ()" {- keep show bindings terse -}
         ,"let ġė_Text = ()" {- keep show bindings terse -} ] )
:}

:{
:def class GhciExt.cmdHlp
     ( \classId -> return $ unlines
       [":redir ġė_Text :info "++classId
       ,"let notInstance = not . Data.List.isPrefixOf \"instance\" in"++
         " putStr $unlines $takeWhile notInstance  $lines ġė_Text"
       ,"let ġė_Text = ()" {- keep show bindings terse -} ] )
     ":class <classId>          -- show class methods" ""
:}

:{
:def defs GhciExt.cmdHlp
     ( \arg ->
       let{ action hlpOption = return $
         let{ cmds = [".","*","x","X","output","redir","grep","find","locate","repeat",
           "inject","strobe","monitor","watch","count","class","defs"] }in
         unlines $map (\c->":"++c++" "++hlpOption) cmds }in
       case arg of{
         "long" -> action "--help";
         "" -> action "-h";
         _ -> return "putStrLn \"usage: :defs [long]\"" } )
     ":defs [long]              -- list user commands (possibly with long help)"
     $unlines $map (\s->' ':s)
       ["General information for all GhciExt commands:"
       ,"- Breakpoints accept an optional argument &<c>. If it is present then"
       ,"  bolean expression <c> limits when anything related to the breakpoint can"
       ,"  run at all (i.e. it can switch off even incrementing of breakpoint hits)."
       ,"- Most arguments may be in quotation marks (to allow spaces in them)."
       ,"- Instead of a regular quotation mark, @x can be used to start a string"
       ,"  argument too. In such a case the argument is finished when character"
       ,"  x appears again after `@x' string. Use any character in place of x."
       ,"- Most commands can have an end-line comment. The comment starts with"
       ,"  this sequence: \"d--\", where d is the argument delimiter; i.e. it"
       ,"  is '\"', the x after '@' or whitespace based on the previous argument."
       ,"- GhciExt calls :init anytime GHCi may have cleared your top level symbols."
       ,"  Redefine :init to reintroduce the symbols you want always available."]
:}

